/*
  File name        : calloc.c
  Date of creation : 11/5/2009
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains macros for debugging

  History          :
  - initial version
*/

/* depend: */
/* cflags: -D_XOPEN_SOURCE=500 */
/* lflags: */

#include <assert.h>
#include <execinfo.h>
#include <malloc.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

extern void __libc_free (void *pt);
extern void *__libc_malloc (size_t size);
extern void *__libc_realloc (void *pt, size_t size);

inline unsigned int nextpow10 (unsigned int x) {
    unsigned int n = 0;
    while (x) {
        n++;
        x = x / 10;
    }
    return (n == 0) ? 1 : n;
}

/* simple fprintf function */

size_t __calloc_fprintf (FILE *fid, const char *fmt, ...)
{
    char buffer[1024 + 1] = { 0 };
    char *str = buffer;

    va_list ap;
    va_start (ap, fmt);
    while (*fmt) {
        char *s;
        int d = 0;
        unsigned int u;
        void *p;
        char c = *fmt++;

        /* copy standard char */
        if (c != '%') {
            *str++ = c;
        } else {
            int t = 0;
            size_t i;

            /* process format char */
            switch (*fmt++) {
            case 'c': /* char */
                c = (char) va_arg (ap, int);
                *str++ = c;
                break;
            case 'd': /* int */
                d = va_arg (ap, int);
                if (d < 0) {
                    *str++ = '-';
                    d = -d;
                }
                t = 1;
                /* fall through */
            case 'u': /* unsigned int */
                u = (t) ? (unsigned int)d : va_arg (ap, unsigned int);
                for (i = nextpow10 (u), s = str; i > 0; i--, s++) {
                    str[i - 1] = '0' + (u % 10);
                    u /= 10;
                }
                str = s;
                break;
            case 'p': /* pointer */
                p = va_arg (ap, void *);
                *str++ = '0';
                *str++ = 'x';
                for (i = sizeof (void *) * 2; i > 0; i--) {
                    char x = (char)(((uintptr_t)p >> (i * 4 - 4)) & 0xf);
                    *str++ = (x > 9) ? 'a' + x - 10 : '0' + x;
                }
                break;
            case 's': /* string */
                s = va_arg (ap, char *);
                while (s && *s)
                    *str++ = *s++;
                break;
            default:
                *str++ = '?';
            }
        }
    }
    va_end (ap);

    /* output string */
    size_t n = str - buffer;
    if (n < sizeof (buffer) - 1) {
        fwrite (buffer, n, 1, fid);
        return n;
    }
    return 0;
}

#undef assert
#define assert(expr)                                                    \
    ((expr)                                                             \
     ? __ASSERT_VOID_CAST (0)                                           \
     : myassert_fail (__STRING(expr), __FILE__, __LINE__, __ASSERT_FUNCTION))

void myassert_fail (char *expr, char *file, int line, const char *func)
{
    __calloc_fprintf (stderr, "assert on %s in %s:%d %s\n", expr, file, line, func);
    abort ();
}

/**
   Debug memory allocation control
*/
int check_memory_allocation_ = -1;

#include "calloc.h"

void sig_handler (int sig) {
    __calloc_fprintf (stdout, "received signal %d\n", sig);
    display_allocation_table ();
    abort ();
}

/**
   Check environment variable MEMORY_ALLOCATION_ to set memory
   allocation control. Possible values are (0, "no"), (1, "yes") or
   (2, "tolerant").
*/
inline int check_memory_allocation (void)
{
    if (check_memory_allocation_ < 0) {
        char *answer = getenv ("MEMORY_ALLOCATION_");
        char test = (answer == NULL) ? 0 : answer[0];
        switch (test) {
        case '1':
            check_memory_allocation_ = (answer[1] == 0) ? 1 : 2;
            break;
        case 'Y': /* yes */
        case 'y':
            check_memory_allocation_ = 1;
            break;
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case 'T': /* tolerant */
        case 't':
            check_memory_allocation_ = 2;
            break;
        default:
            check_memory_allocation_ = 0;
        }
        if (check_memory_allocation_) {
            atexit (display_allocation_table);
            signal (SIGSEGV, sig_handler);
            signal (SIGILL, sig_handler);
            signal (SIGBUS, sig_handler);
            __calloc_fprintf (stdout, "check_memory_allocation_: %d\n", check_memory_allocation_);
        }
    }
    return check_memory_allocation_;
}

/**
   Code reference (file and line)
*/
typedef struct {
    char *file; /**< Filename */
    int line; /**< Line number */
} reference_t;

/**
   Memory allocation structure
*/
typedef struct {
    void *pt; /**< Memory pointer */
    size_t size; /**< Effective size */
    reference_t begin; /**< Allocation reference */
    reference_t end; /**< Free reference */
    int free; /**< State: 0 active, 1 freed, 1 reallocated */
} mem_block_t;

/**
   Global memory allocation table
*/
mem_block_t *mem_blocks = NULL;

/**
   Number of elements in the memory allocation table
*/
int nb_mem_blocks = 0;

/**
   Increment allocation step of the memory allocation table
*/
#define MEM_BLOCKS 1024

/**
   Number of extra bytes which will be added before and after the
   memory block to protect it
*/
#define PROTECTION 16

/**
   Protection mark for postmortem control
*/
#define MARK 42

/**
   Thread protection mutex
*/
static pthread_mutex_t check_memory_allocation_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Macros have to be undefined */
#undef calloc
#undef free
#undef malloc
#undef realloc

/**
   Insert a new reference into memory allocation table.

   @param pt effective pointer
   @param size user pointer size
   @param file file name where allocation has been called
   @param line line number where allocation has been called
   @return allocation reference
*/
mem_block_t *newallocation (void *pt, size_t size, const char *file, const int line)
{
    __calloc_fprintf (stdout, "newallocation: %p %u %s %d\n", pt, size, file, line);

    mem_block_t *ref;

    /* thread protection begin */
    pthread_mutex_lock (&check_memory_allocation_mutex);

    /* memory table allocation */
    if (nb_mem_blocks % MEM_BLOCKS == 0) {

        mem_block_t *mb = (mem_block_t *) __libc_malloc ((nb_mem_blocks + MEM_BLOCKS) * sizeof (mem_block_t));
        if (mb == NULL) {
            pthread_mutex_unlock (&check_memory_allocation_mutex);
            return NULL;
        }
        memcpy (mb, mem_blocks, nb_mem_blocks * sizeof (mem_block_t));
        memset (mb + nb_mem_blocks, 0, MEM_BLOCKS * sizeof (mem_block_t));
        __libc_free (mem_blocks);
        mem_blocks = mb;
    }
    ref = mem_blocks + nb_mem_blocks++;

    /* thread protection end */
    pthread_mutex_unlock (&check_memory_allocation_mutex);

    /* get file name from backtrace */
    char **name = NULL;
    if (!file) {
        void *array[4];
        int saved_value = check_memory_allocation_;
        check_memory_allocation_ = 0;
        int nb = backtrace (array, 4);
        if (nb > 0) {
            name = backtrace_symbols (array, nb);
            file = name[nb - 1];
        }
        check_memory_allocation_ = saved_value;
    } else
        if (file[0] == '#') file++;

    /* fulfill memory table allocation */
    ref->pt = pt;
    ref->size = size;
    ref->begin.file = (char *) __libc_malloc (strlen (file) + 1);
    if (ref->begin.file)
        strcpy (ref->begin.file, file);
    if (name) __libc_free (name);
    ref->begin.line = line;
    ref->free = 0;

    return ref;
}

/**
   Look for an allocation block into memory allocation table.

   @param pt pointer
   @return allocation reference
*/
mem_block_t *lookforallocation (void *pt)
{
    int i;
    mem_block_t *ref = NULL;

    if (pt == NULL)
        return NULL;

    for (i = 0; i < nb_mem_blocks; i++)
        if (mem_blocks[i].pt == pt - PROTECTION) {
            ref = mem_blocks + i;
            break;
        }

    return ref;
}

/**
   Check memory block consistancy

   @param pt effective pointer
   @param size user pointer size
   @return status (1 for consistant)
*/
int check_block_consistancy (void *pt, size_t size)
{
    int state = 0;
    int i;

    for (i = 0; (i < PROTECTION) && (state != (2 | 4)); i++) {
        if (((char *)pt)[PROTECTION - 1 - i] != MARK) state |= 2;
        if (((char *)pt)[PROTECTION + size + i] != MARK) state |= 4;
    }

    return state;
}

void *mymalloc (size_t size, const char *file, const int line)
{
    void *pt;

    /* standard allocation */
    if (check_memory_allocation () == 0) {
        pt = __libc_malloc (size);
        return pt;
    }

    /* pointer allocation */
    pt = (void *) __libc_malloc (2 * PROTECTION + size);
    assert (pt);
    memset (pt, MARK, PROTECTION);
    memset (pt + PROTECTION + size, MARK, PROTECTION);

    /* update memory table allocation */
    newallocation (pt, size, file, line);

    return pt + PROTECTION;
}

void *mycalloc (size_t size, const char *file, const int line)
{
    void *pt;

    /* standard allocation */
    if (check_memory_allocation () == 0) {
        pt = (void *) __libc_malloc (size);
        if (pt) memset (pt, 0, size);
        return pt;
    }

    /* pointer allocation */
    pt = (void *) __libc_malloc (2 * PROTECTION + size);
    assert (pt);
    memset (pt, MARK, PROTECTION);
    memset (pt + PROTECTION, 0, size);
    memset (pt + PROTECTION + size, MARK, PROTECTION);

    /* update memory table allocation */
    newallocation (pt, size, file, line);

    return pt + PROTECTION;
}

void myfree (void *pt, const char *file, const int line)
{
    mem_block_t *ref;

    /* standard free */
    if (check_memory_allocation () == 0) {
        __libc_free (pt);
        return;
    }

    /* get file name from backtrace */
    char **name = NULL;
    if (!file) {
        void *array[3];
        int saved_value = check_memory_allocation_;
        check_memory_allocation_ = 0;
        int nb = backtrace (array, 3);
        if (nb > 0) {
            name = backtrace_symbols (array, nb);
            file = name[nb - 1];
        }
        check_memory_allocation_ = saved_value;
    } else
        if (file[0] == '#') file++;

    /* check null free */
    if (!pt) {
        if (check_memory_allocation () > 1) {
            __calloc_fprintf (stderr, "free null pointer");
            if (file) {
                __calloc_fprintf (stderr, " at %s", file);
                if (line > 0) __calloc_fprintf (stderr, ":%d", line);
            }
            __calloc_fprintf (stderr, "\n");
        } else
            assert (pt != NULL);
        return;
    }

    /* look for pointer in memory allocation table */
    ref = lookforallocation (pt);
    if (!ref) {
        if (check_memory_allocation () > 1) {
            __calloc_fprintf (stderr, "try to free an unknown pointer %p", pt);
            if (file) {
                __calloc_fprintf (stderr, " at %s", file);
                if (line > 0) __calloc_fprintf (stderr, ":%d", line);
            }
            __calloc_fprintf (stderr, "\n");
        } else
            assert (ref != NULL);
        return;
    }

    /* check memory block consistancy */
    int state = 1;
    if (check_memory_allocation () == 1)
        assert (!check_block_consistancy (ref->pt, ref->size));
    else if ((state = check_block_consistancy (ref->pt, ref->size)) != 0) {
        __calloc_fprintf (stderr, "unconsistant (%c%c) pointer %p allocated",
                 (state & 2) ? 'B' : '.', (state & 4) ? 'E' : '.', pt);
        if (ref->begin.file) {
            __calloc_fprintf (stderr, " in %s", ref->begin.file);
            if (ref->begin.line > 0) __calloc_fprintf (stderr, ":%d", ref->begin.line);
        } else
            __calloc_fprintf (stderr, " somewhere");
        __calloc_fprintf (stderr, " and freed");
        if (file) {
            __calloc_fprintf (stderr, " in %s", file);
            if (line > 0) __calloc_fprintf (stderr, ":%d", line);
        } else
            __calloc_fprintf (stderr, " later");
        __calloc_fprintf (stderr, "\n");
    } else {
        state = 1;
    }

    /* free pointer */
    __libc_free (ref->pt);

    /* fulfill memory allocation table */
    ref->free = state;
    ref->end.file = (char *) __libc_malloc (strlen (file) + 1);
    if (ref->end.file)
        strcpy (ref->end.file, file);
    if (name) __libc_free (name);
    ref->end.line = line;
}

void *myrealloc (void *pt, size_t size, const char *file, const int line)
{
    mem_block_t *ref;

    /* standard realloc */
    if (check_memory_allocation () == 0) {
        pt = (void *) __libc_realloc (pt, size);
        return pt;
    }

    /* first pointer allocation */
    if (pt == NULL) {

        /* pointer allocation */
        pt = (void *) __libc_malloc (2 * PROTECTION + size);
        assert (pt);
        memset (pt, MARK, PROTECTION);
        memset (pt + PROTECTION, 0, size);
        memset (pt + PROTECTION + size, MARK, PROTECTION);

        /* update memory table allocation */
        ref = newallocation (pt, size, file, line);

    }
    /* real reallocation */
    else {

        /* get file name from backtrace */
        char **name = NULL;
        if (!file) {
            void *array[4];
            int saved_value = check_memory_allocation_;
            check_memory_allocation_ = 0;
            int nb = backtrace (array, 4);
            if (nb > 0) {
                name = backtrace_symbols (array, nb);
                file = name[nb - 1];
            }
            check_memory_allocation_ = saved_value;
        } else
            if (file[0] == '#') file++;

        /* look for pointer in memory allocation table */
        ref = lookforallocation (pt);
        if (ref == NULL) {
            __calloc_fprintf (stderr, "try to reallocate an unknown pointer %p at %s",
                     pt, file);
            if (line > 0) __calloc_fprintf (stderr, ":%d", line);
            __calloc_fprintf (stderr, "\n");
        }

        assert (ref != NULL);

        /* pointer allocation */
        void *npt = (void *) __libc_malloc (2 * PROTECTION + size);
        assert (npt);
        memset (npt, MARK, PROTECTION);
        memset (npt + PROTECTION, 0, size);
        memset (npt + PROTECTION + size, MARK, PROTECTION);

        /* copy old data in new buffer */
        memcpy (npt + PROTECTION, ref->pt + PROTECTION,
                (ref->size < size) ? ref->size : size);

        /* check memory block consistancy */
        int state = 1;
        if (check_memory_allocation () == 1)
            assert (!check_block_consistancy (ref->pt, ref->size));
        else if ((state = check_block_consistancy (ref->pt, ref->size)) != 0) {
            __calloc_fprintf (stderr, "unconsistant (%c%c) pointer %p allocated in %s",
                     (state & 2) ? 'B' : '.', (state & 4) ? 'E' : '.',
                     pt, ref->begin.file);
            if (ref->begin.line > 0) __calloc_fprintf (stderr, ":%d", ref->begin.line);
            __calloc_fprintf (stderr, " and reallocated in %s", file);
            if (line > 0) __calloc_fprintf (stderr, ":%d", line);
            __calloc_fprintf (stderr, "\n");
        } else {
            state = 1;
        }

        /* free old pointer */
        __libc_free (ref->pt);
        ref->free = -state;

        /* swap new pointer */
        ref->pt = npt;
        ref->size = size;

        /* update last allocation reference*/
        if (ref->begin.file) {
            __libc_free (ref->begin.file);
        }
        ref->begin.file = (char *) __libc_malloc (strlen (file) + 1);
        if (ref->begin.file)
            strcpy (ref->begin.file, file);
        if (name) __libc_free (name);
        ref->begin.line = line;
    }

    return ref->pt + PROTECTION;
}

void display_allocation_table (void)
{
    int i;

    /* header */
    __calloc_fprintf (stdout, "allocation table summary\n"
             "------------------------\n"
             "stat | pointer  | size (B) | reference (alloc free)\n");

    for (i = 0; i < nb_mem_blocks; i++) {

        /* generate status */
        char status[6] = { 'X', '.', '.', ' ', ' ', 0};
        if (mem_blocks[i].free > 0) {
            status[0] = 'F';
            if (mem_blocks[i].free & 2) status[1] = 'B';
            if (mem_blocks[i].free & 4) status[2] = 'E';
        } else {
            status[0] = (mem_blocks[i].free == 0) ? 'A' : 'R';

            /* check consistancy */
            int state =
                check_block_consistancy (mem_blocks[i].pt, mem_blocks[i].size);
            int offset = 0;
            if (mem_blocks[i].free < 0) {
                if (-mem_blocks[i].free & 2) status[1] = 'B';
                if (-mem_blocks[i].free & 4) status[2] = 'E';
                offset = 2;
            }
            status[offset + 1] = (state & 2) ? 'B' : '.';
            status[offset + 2] = (state & 4) ? 'E' : '.';
        }

        /* file reference */
        __calloc_fprintf (stdout, "%s %p %d %s", status, mem_blocks[i].pt,
                 (int)mem_blocks[i].size, mem_blocks[i].begin.file);
        if (mem_blocks[i].begin.line > 0)
            __calloc_fprintf (stdout, ":%d", mem_blocks[i].begin.line);
        if (mem_blocks[i].free > 0)
            __calloc_fprintf (stdout, " %s", mem_blocks[i].end.file);
        if (mem_blocks[i].end.line > 0)
            __calloc_fprintf (stdout, ":%d", mem_blocks[i].end.line);

        __calloc_fprintf (stdout, "\n");
    }
}

/* malloc hooks*/

void free (void *pt) {
    return myfree (pt, NULL, -1);
}

void *malloc (size_t size) {
    return mymalloc (size, NULL, -1);
}

void *realloc (void *pt, size_t size) {
    return myrealloc (pt, size, NULL, -1);
}

void *calloc (size_t nmemb, size_t size) {
    return mycalloc (nmemb * size, NULL, -1);
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
